import pandas as pd
from dataclasses import dataclass
from typing import TypedDict, List, Optional
from athina.interfaces.data import DataPoint


class LlmEvalResultMetric(TypedDict):
    """
    Represents the LLM evaluation result metric.
    """

    id: str
    value: float


class LlmEvalResult(TypedDict):
    """
    Represents the LLM evaluation result.
    """

    name: str
    display_name: str
    data: DataPoint
    failure: int
    reason: str
    runtime: int
    model: str
    metric: Optional[LlmEvalResultMetric]

@dataclass
class BatchRunResult:
    """
    Represents the result of a batch run of LLM evaluation.
    """

    eval_request_id: str
    eval_results: List[LlmEvalResult]

    def to_df(self):
        """
        Converts the batch run result to a Pandas DataFrame.
        """
        pd.set_option('display.max_colwidth', 500)
        df = pd.DataFrame(self.eval_results)

        # Normalize the 'data' column
        results_df = df.drop(columns=['data', 'name', 'metric']).rename(columns={'failure': 'failed', 'name': 'eval','reason': 'grade_reason'})
        data_normalized = pd.json_normalize(df['data'])
        metric_normalized = pd.json_normalize(df['metric']).rename(columns={'id': 'metric_id', 'value': 'metric_value'})

        # Concatenate the normalized data with the original DataFrame (excluding the 'data' column)
        df = pd.concat([data_normalized, results_df, metric_normalized], axis=1)
        return df


class EvalPerformanceReport(TypedDict):
    """
    Represents the performance metrics for an evaluation.
    """

    true_positives: int
    false_positives: int
    true_negatives: int
    false_negatives: int
    accuracy: float
    precision: float
    recall: float
    f1_score: float
    runtime: int
    dataset_size: int



from enum import Enum


class AthinaEvalTypeId(Enum):
    CONTEXT_CONTAINS_ENOUGH_INFORMATION = "Ccei"
    DOES_RESPONSE_ANSWER_QUERY = "Draq"
    FAITHFULNESS = "Irftc"
    CUSTOM = "Custom"
    SUMMARY_ACCURACY = "SummaryAccuracy"



from enum import Enum
import pprint
import time
import traceback
from typing import List, Optional
from athina.interfaces.model import Model
from athina.interfaces.result import LlmEvalResult, LlmEvalResultMetric, BatchRunResult
from athina.loaders.summary_loader import SummaryDataPoint
from athina.metrics.metric_type import MetricType
from ..llm_evaluator import LlmEvaluator
from ..eval_type import AthinaEvalTypeId
from ..example import FewShotExample
from athina.llms.question_answerer import QuestionAnswerer
from athina.llms.question_answerer_bulk import QuestionAnswererBulk
from athina.llms.question_answerer_cot import QuestionAnswererChainOfThought
from athina.llms.question_generator import QuestionGenerator

class SummaryAccuracy(LlmEvaluator):
    """
    This evaluator can be configured with custom examples and instructions.
    """
    questions: List[str] = []

    def __init__(
        self,
        questions: Optional[List[str]] = None,
        n_questions: int = 10,
        model: str = "gpt-4-1106-preview",
        question_answerer: Optional[QuestionAnswerer] = None,
        metrics: List[MetricType] = [
            MetricType.AGREEMENT_SCORE,
            MetricType.CONTRADICTION_SCORE,
            MetricType.HALLUCINATION_SCORE
        ],
    ):
        """
        Initialize the evaluator with given parameters.

        Args:
        - loader: An instance of SummarizationLoader.
        - n_questions: Number of questions to generate for summaries.
        - llm_model: Language model to be used.
        - metrics: List of metrics for evaluation.
        """

        # Intialize LLMs
        self._model = model
        self.n_questions = n_questions
        if questions is not None:
            self.questions = questions
        self.question_generator = QuestionGenerator(
            self._model, n_questions
        )
        if question_answerer is None:
            self.question_answerer = QuestionAnswererBulk(model=self._model)
        else:
            self.question_answerer = question_answerer
        self.n_instances = 0
        self.metrics: List[MetricType] = metrics
        self.label_counts = {}
        for metric in metrics:
            setattr(self, f"{metric}_scores", {})

    @property
    def name(self):
        return AthinaEvalTypeId.CUSTOM.value
        
    @property
    def metric_id(self) -> str:
        return MetricType.AGREEMENT_SCORE.value
        
    @property
    def display_name(self):
        return "Summary Accuracy"
        
    @property
    def default_model(self):
        return Model.GPT35_TURBO.value
        
    @property
    def required_args(self):
        return ["document", "response"]
        
    @property
    def examples(self):
        return []
        
    def is_failure(self) -> bool:
        return False
    
    def reason(self) -> str:
        return ""


    def _evaluate(self, **instance) -> LlmEvalResult:
        """
        Run the LLM evaluator.
        """
        start_time = time.time()

        # Validate that correct args were passed
        self._validate_args(**instance)

        summary_datapoint = SummaryDataPoint(**instance)

        # Run the Summary Accuracy evaluator
        summary_eval_result = self._evaluate_element(summary_datapoint)

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        
        llm_eval_result = LlmEvalResult(
            name=self.name,
            display_name=self.display_name,
            data=SummaryDataPoint(**instance),
            failure=self.is_failure(),
            reason=self.reason(),
            runtime=eval_runtime_ms,
            model=self._model,
            metric={
                "id": self.metric_id,
                "value": summary_eval_result[self.metric_id],
            },
        )

        return {k: v for k, v in llm_eval_result.items() if v is not None}
    

    def _evaluate_element(self, instance: SummaryDataPoint):
        """Evaluate an instance for hallucination."""
        try:
            document = instance["document"]
            summary = instance["response"]
            if "label" in instance:
                label = instance["label"]
            else:
                label = "overall"

            # Generate questions based on summary
            if self.questions is None or len(self.questions) == 0:
                self.questions = self.question_generator.generate(summary)

            answers_doc = self.question_answerer.answer(questions=self.questions, context=document)[1]
            answers_sum = self.question_answerer.answer(questions=self.questions, context=summary)[1]
            metric_results = {}
            # Compute metrics
            if answers_doc is None or answers_sum is None or self.questions is None:
                raise Exception("Validation error - unable to generate answers")
            else:
                for metric in self.metrics:
                    metric_name = metric.value
                    metric_class = metric.get_class()
                    metric_result, explanation = metric_class.compute(
                        answers_doc, answers_sum, self.questions, self.n_questions
                    )
                    metric_results[metric_name] = metric_result
                    metric_results[f"reason_{metric_name}"] = explanation
                    self.update_metric_aggregated_score(metric_name, label, metric_result)
                self.n_instances = self.n_instances + 1
                self.label_counts[label] = self.label_counts.get(label, 0) + 1
            return {
                "questions": self.questions,
                "answers_doc": answers_doc,
                "answers_sum": answers_sum,
                "label": label,
                **metric_results,
            }
        except Exception as e:
            print("Exception in _evaluate_element", e)
            traceback.print_exc()
            raise e

    def update_metric_aggregated_score(self, metric, label, aggr_score):
        """Update the aggregated score for a specific metric and label."""
        metric_aggregated_scores = getattr(self, f"{metric}_scores", {})
        current_score = metric_aggregated_scores.get(label, 0)
        metric_aggregated_scores[label] = current_score + aggr_score
        setattr(self, f"{metric}_scores", metric_aggregated_scores)

    def get_metric_aggr(self, metric, label):
        """Compute the average scores based on the provided score dictionary."""
        metric_aggr = getattr(self, f"{metric}_scores", {})
        return metric_aggr.get(label, None)

    def get_average_scores(self, score_dict):
        """Compute average scores for a metric"""
        avg_scores = {}
        sum_score = 0
        n_instances = 0
        for label_type, total_score in score_dict.items():
            avg_scores[label_type] = total_score / self.label_counts[label_type]
            sum_score = sum_score + total_score
            n_instances = n_instances + self.label_counts[label_type]
        avg_scores["overall"] = sum_score / n_instances
        return avg_scores

    def compute_average_scores(self):
        """Compute average scores for each metric."""
        avg_scores = {}
        for metric in self.metrics:
            scores = getattr(self, f"{metric}_scores")
            avg_score = self.get_average_scores(scores)
            avg_scores[metric] = avg_score
        return avg_scores

    


from enum import Enum
from .agreement_score import AgreementScore
from .hallucination_score import HallucinationScore
from .contradiction_score import ContradictionScore

class MetricType(Enum):
    AGREEMENT_SCORE = 'agreement_score'
    HALLUCINATION_SCORE = 'hallucination_score'
    CONTRADICTION_SCORE = 'contradiction_score'

    def get_class(self):
        if self == MetricType.AGREEMENT_SCORE:
            return AgreementScore
        if self == MetricType.HALLUCINATION_SCORE:
            return HallucinationScore
        if self == MetricType.CONTRADICTION_SCORE:
            return ContradictionScore
        else:
            raise NotImplementedError(f"Metric type {self} not implemented.")



from typing import List, Optional
from ..llm_evaluator import LlmEvaluator
from ..eval_type import AthinaEvalTypeId
from ..example import FewShotExample


class CustomLlmEval(LlmEvaluator):
    """
    This evaluator can be configured with custom examples and instructions.
    """

    _display_name: str = None
    _metric_id: str = None
    _model: str = None
    _required_args: List[str] = None
    _examples: List[FewShotExample] = None
    _system_message_template: Optional[str] = None
    _user_message_template: Optional[str] = None

    def __init__(
        self,
        display_name: str = None,
        metric_id: str = None,
        model: str = None,
        required_args: List[str] = [],
        examples: List[FewShotExample] = [],
        system_message_template: Optional[str] = None,
        user_message_template: Optional[str] = None,
        **kwargs
    ):
        if display_name is None:
            raise ValueError("display_name is not defined")
        if model is None:
            raise ValueError("model is not defined")

        self._display_name = display_name
        self._metric_id = metric_id
        self._model = model
        self._required_args = required_args
        self._examples = examples

        if system_message_template is not None:
            self._system_message_template = system_message_template

        if user_message_template is not None:
            self._user_message_template = user_message_template

        super().__init__(
            model=self._model,
            system_message_template=self._system_message_template,
            user_message_template=self._user_message_template,
            **kwargs
        )

    @property
    def name(self):
        return AthinaEvalTypeId.CUSTOM.value
        
    @property
    def metric_id(self) -> str:
        return self._metric_id
        
    @property
    def display_name(self):
        return self._display_name
        
    @property
    def default_model(self):
        return self._model
        
    @property
    def required_args(self):
        return self._required_args
        
    @property
    def examples(self):
        return self._examples
        



from abc import ABC, abstractmethod
import time
from typing import List, Optional
from concurrent.futures import ThreadPoolExecutor, as_completed
from athina.interfaces.result import LlmEvalResult, LlmEvalResultMetric, BatchRunResult
from athina.interfaces.athina import AthinaExperiment
from athina.interfaces.model import Model
from athina.llms.openai_service import OpenAiService
from athina.helpers.logger import logger
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.interfaces.data import DataPoint
from athina.services.athina_api_service import AthinaApiService
from .example import FewShotExample


class LlmEvaluator(ABC):
    llm_service: OpenAiService
    grading_criteria: str
    _model: str
    _experiment: Optional[AthinaExperiment] = None
    _system_message_template: Optional[str] = None
    _user_message_template: Optional[str] = None

    TEMPERATURE = 0.0

    RETURN_FORMAT_INSTRUCTIONS = """
    You MUST return a JSON object with the following fields: 
    - result: Result must be either 'Pass' or 'Fail'.
    - explanation: An explanation of why the result is Pass or Fail.
    - score: (Optional) Use the scoring criteria specified.
    """

    DEFAULT_SYSTEM_MESSAGE_TEMPLATE = f""" 
    ### INSTRUCTIONS ###
    You are an expert at evaluating chatbot responses, according to some grading criteria.

    If it passes the grading criteria, then your result is Pass, otherwise it is Fail.
    
    """

    DEFAULT_USER_MESSAGE_TEMPLATE = """
    ### GRADING CRITERIA ###
    {grading_criteria}

    ### EXAMPLES ###
    {examples}

    ### RESPONSE TO EVALUATE ###
    {response}
    """

    EXAMPLES: FewShotExample = []

    def __init__(
        self,
        model: Optional[str] = None,
        grading_criteria: Optional[str] = None,
        system_message_template: Optional[str] = None,
        user_message_template: Optional[str] = None,
    ):
        self.llm_service = OpenAiService()
        self.grading_criteria = grading_criteria if grading_criteria else ""
        if model is None:
            self._model = self.default_model
        elif not Model.is_supported(model):
            raise ValueError(f"Unsupported model: {model}")
        else:
            self._model = model

        # Initialize message templates
        if system_message_template is None:
            self._system_message_template = (
                self.DEFAULT_SYSTEM_MESSAGE_TEMPLATE + self.RETURN_FORMAT_INSTRUCTIONS
            )
        else:
            self._system_message_template = system_message_template

        if user_message_template is None:
            self._user_message_template = self.DEFAULT_USER_MESSAGE_TEMPLATE
        else:
            self._user_message_template = user_message_template

    # Abstract properties
    @property
    @abstractmethod
    def name(self) -> str:
        """A unique name identifier for the evaluator."""
        pass

    @property
    @abstractmethod
    def display_name(self) -> str:
        """A display name for the evaluator."""
        pass

    @property
    @abstractmethod
    def metric_id(self) -> str:
        """The metric computed by the evaluator."""
        pass

    @property
    @abstractmethod
    def required_args(self):
        """A list of required arguments for the evaluator."""
        pass

    @property
    @abstractmethod
    def default_model(self):
        """The default model for the evaluator."""
        pass

    @property
    @abstractmethod
    def examples(self):
        """A list of examples for the evaluator."""
        pass

    # Common methods
    def _examples_str(self) -> str:
        return "\n".join([str(example) for example in self.examples])

    def _system_message(self) -> str:
        return self._system_message_template

    def _user_message(self, **kwargs) -> str:
        return self._user_message_template.format(
            examples=self._examples_str(),
            grading_criteria=self.grading_criteria,
            **kwargs,
        )

    def _prompt_messages(self, **kwargs) -> List[dict]:
        return [
            {
                "role": "system",
                "content": self._system_message(),
            },
            {
                "role": "user",
                "content": self._user_message(**kwargs),
            },
        ]

    def _validate_args(self, **kwargs) -> None:
        for arg in self.required_args:
            if arg not in kwargs:
                raise ValueError(f"Missing required argument: {arg}")

    def _evaluate(self, **kwargs) -> LlmEvalResult:
        """
        Run the LLM evaluator.
        """
        start_time = time.time()

        # Validate that correct args were passed
        self._validate_args(**kwargs)

        # Construct Prompt
        messages = self._prompt_messages(**kwargs)

        # Run the LLM Completion

        chat_completion_response_json: dict = self.llm_service.json_completion(
            model=self._model,
            messages=messages,
            temperature=self.TEMPERATURE,
        )

        try:
            metric = None
            result = chat_completion_response_json["result"]
            explanation = chat_completion_response_json["explanation"]
            failure = bool(result == "Fail")
            if "score" in chat_completion_response_json:
                score = chat_completion_response_json["score"]
                metric = LlmEvalResultMetric(id=self.metric_id, value=score)
            else:
                metric = LlmEvalResultMetric(id="failed", value=float(failure))

        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        llm_eval_result = LlmEvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            failure=failure,
            reason=explanation,
            runtime=eval_runtime_ms,
            model=self._model,
            metric=metric,
        )
        return {k: v for k, v in llm_eval_result.items() if v is not None}

    def configure_experiment(self, experiment: AthinaExperiment):
        """Configured metadata parameters to log an experiment to Athina"""
        self._experiment = experiment
        return self

    def run(self, **kwargs) -> BatchRunResult:
        """
        Run the LLM evaluator, and log results to Athina.
        """
        # Log usage to Athina for analytics
        AthinaApiService.log_usage(eval_name=self.name, run_type="single")

        # Create eval request
        eval_request_id = AthinaLoggingHelper.create_eval_request(
            eval_name=self.name, request_data=kwargs, request_type="single"
        )

        # Log experiment
        if self._experiment:
            AthinaLoggingHelper.log_experiment(
                eval_request_id=eval_request_id,
                experiment=self._experiment,
            )

        eval_result = self._evaluate(**kwargs)

        # Log evaluation results to Athina
        AthinaLoggingHelper.log_eval_results(
            eval_request_id=eval_request_id,
            eval_results=[eval_result],
        )

        return BatchRunResult(
            eval_request_id=eval_request_id,
            eval_results=[eval_result],
        )

    def _validate_batch_args(self, data: List[DataPoint]) -> bool:
        """
        Validates that each entry in the batch has all the required arguments.
        """
        for i, entry in enumerate(data):
            for arg in self.required_args:
                if arg not in entry:
                    raise ValueError(
                        f"Data at index {i} is missing required argument: {arg}"
                    )
        return True

    def _run_batch_generator_async(
        self, data: List[DataPoint], max_parallel_evals: int
    ):
        with ThreadPoolExecutor(max_workers=max_parallel_evals) as executor:
            # Submit all tasks to the executor and store them with their original index
            future_to_index = {
                executor.submit(self._evaluate, **entry): i
                for i, entry in enumerate(data)
            }

            # Create a list to store results in the original order
            results = [None] * len(data)

            for future in as_completed(future_to_index):
                index = future_to_index[future]
                try:
                    results[index] = future.result()
                except Exception as e:
                    entry = data[index]
                    logger.error(f"Error evaluating entry {entry}: {e}")
                    results[index] = None

            return results

    def _run_batch_generator(self, data: List[DataPoint]):
        """
        Generator function for running a batch of evaluations.
        Iterates over a dataset, and runs the evaluator on each entry.
        """
        for entry in data:
            try:
                yield self._evaluate(**entry)
            except Exception as e:
                logger.error(f"Error evaluating entry {entry}: {e}")
                yield None

    def run_batch(
        self, data: List[DataPoint], max_parallel_evals: int = 1
    ) -> BatchRunResult:
        """
        Runs the evaluator on a batch of data.
        """

        # Create eval request
        eval_request_id = AthinaLoggingHelper.create_eval_request(
            eval_name=self.name, request_data={"data": data}, request_type="batch"
        )

        # Log usage to Athina for analytics
        AthinaApiService.log_usage(eval_name=self.name, run_type="batch")

        # Log experiment
        if self._experiment is not None:
            AthinaLoggingHelper.log_experiment(
                eval_request_id=eval_request_id,
                experiment=self._experiment,
            )

        # Validate the dataset against the required args
        self._validate_batch_args(data)

        # Run the evaluations
        if max_parallel_evals > 1:
            eval_results = self._run_batch_generator_async(data, max_parallel_evals)
        else:
            eval_results = list(self._run_batch_generator(data))

        # Log evaluation results to Athina
        AthinaLoggingHelper.log_eval_results(
            eval_request_id=eval_request_id,
            eval_results=eval_results,
        )

        return BatchRunResult(
            eval_request_id=eval_request_id,
            eval_results=eval_results,
        )
